@@ -12,7 +12,7 @@ if(isprint("abc\000abc") || isprint("abc\001abc") || !isprint("barra"))
 
 
 use strict;
-our $VERSION = '2.2';
+our $VERSION = '3.0';
 use CGI qw(escapeHTML);
 use CGI::Cookie;
 require Exporter;
@@ -34,6 +34,9 @@ our %defaults = (
 
 our $have_gd_graph = eval { require GD::Graph::lines; };
 
+our $redo_screen_js =
+  "this.form.redo_screen.value=1; this.form.submit();";
+
 sub new {
     my $type = shift;
     my ($class) = ref($type) || $type;
@@ -49,41 +52,51 @@ sub new {
 			['forced','0'],
 			['single','size'],
 			['single','maxsize'],
-			['single','submit']]);
+			['single','submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('hidetext',\&QWizard::Generator::HTML::do_entry,
 		       [['single','name'],
 			['default'],
 			['forced','1'],
 			['single','size'],
 			['single','maxsize'],
-			['single','submit']]);
+			['single','submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('textbox',\&QWizard::Generator::HTML::do_textbox,
 		       [['default'],
 			['single', 'width'],
 			['single', 'size'],
 			['single', 'height'],
-			['single', 'submit']]);
+			['single', 'submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('checkbox',\&QWizard::Generator::HTML::do_checkbox,
 		       [['multi','values'],
 			['default'],
-			['single', 'submit']]);
+			['single', 'submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('multi_checkbox',
 		       \&QWizard::Generator::HTML::do_multicheckbox,
 		       [['multi','default'],
 			['values,labels'],
-			['single','submit']]);
+			['single','submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('menu',
 		       \&QWizard::Generator::HTML::do_menu,
 		       [['values,labels', "   "],
 			['default'],
 			['single','submit'],
+			['single','refresh_on_change'],
 			['single', 'name']]);
     $self->add_handler('radio',
 		       \&QWizard::Generator::HTML::do_radio,
 		       [['values,labels'],
 			['default'],
 			['single','submit'],
-			['single','name']]);
+			['single','refresh_on_change'],
+			['single','name'],
+			['single','icons'],
+			['single','noiconpadding'],
+		       ]);
     $self->add_handler('label',
 		       \&QWizard::Generator::HTML::do_label,
 		       [['multi','values']]);
@@ -102,6 +115,9 @@ sub new {
 		       \&QWizard::Generator::HTML::do_table,
 		       [['norecurse','values'],
 			['norecurse','headers']]);
+    $self->add_handler('bar',
+		       \&QWizard::Generator::HTML::do_bar,
+		       [['norecurse','values']]);
     $self->add_handler('graph',
 		       \&QWizard::Generator::HTML::do_graph,
 		       [['norecurse','values'],
@@ -167,17 +183,15 @@ sub init_screen {
 # html always waits
 sub wait_for {
   my ($self, $wiz, $next, $p) = @_;
-  print "<P><input type=submit value=\"" . escapeHTML($next) . "\">\n";
-  if ($self->qwparam('allow_refresh') || $p->{'allow_refresh'}) {
-      print "<input type=submit name=redo_screen value=\"Refresh Screen\">\n";
-  }
   print $self->{'cgi'}->end_form();
+  print "</tr>\n" if (!exists($self->{'cssurl'}));
+  $self->close_div_or_table(); # end for <div class="qwizard"> in start_primaries
   return 1;
 }
 
 sub do_css {
     my ($self, $class, $name, $noidstr) = @_;
-    if ($self->{'cssurl'}) {
+    if (exists($self->{'cssurl'})) {
 	my $idstr = '';
 	$idstr = $class if (!$noidstr);
 	return " class=\"$class\" id=\"$idstr$name\" ";
@@ -185,20 +199,93 @@ sub do_css {
     return "";
 }
 
+sub open_div_or_table {
+    my $self = shift;
+    if (exists($self->{'cssurl'})) {
+	print "<div class=\"" . $_[0] . "\">\n";
+    } else {
+	print "<table $_[1]>\n";
+    }
+}
+
+sub close_div_or_table {
+    my $self = shift;
+    if (exists($self->{'cssurl'})) {
+	print "</div>\n";
+    } else {
+	print "</table>\n";
+    }
+}
+
+sub open_div_or_tr {
+    my $self = shift;
+    if (exists($self->{'cssurl'})) {
+	print "<div class=\"" . $_[0] . "\">\n";
+    } else {
+	print "<tr $_[1]>\n";
+    }
+}
+
+sub open_span_or_td {
+    my $self = shift;
+    if (exists($self->{'cssurl'})) {
+	print "<span class=\"" . $_[0] . "\">\n";
+    } else {
+	print "<td valign=\"top\" $_[1]>\n";
+    }
+}
+
+sub close_span_or_td {
+    my $self = shift;
+    if (exists($self->{'cssurl'})) {
+	print "</span>\n";
+    } else {
+	print "</td>\n";
+    }
+}
+
+sub close_div_or_tr {
+    my $self = shift;
+    if (exists($self->{'cssurl'})) {
+	print "</div>\n";
+    } else {
+	print "</tr>\n";
+    }
+}
+
+sub get_image {
+    my ($self, $img, $css, $name) = @_;
+
+    return "<img" . $self->do_css($css,$name) .
+      " src=\"" . $self->{'imgpath'} . escapeHTML($img) . "\"> ";
+}
+
 sub do_question {
     my ($self, $q, $wiz, $p, $text, $qcount) = @_;
     return if (!$text && $q->{'type'} eq 'hidden');
+    my $padtext;
+    $padtext=" style=\"padding-left: 2em;\"" if ($q->{'indent'});
     print "  <tr" . $self->do_css('qwquestion',$q->{'name'}) . ">";
     print "<td" . $self->do_css('qwquestiontext',$q->{'name'}) . 
-      " valign=top>\n";
+      " valign=top $padtext>\n";
+    $text = QWizard::Generator::remove_accelerator($text);
     if ($q->{'helptext'}) {
 	print $wiz->make_help_link($p, $qcount),
-	  escapeHTML($text), "</a>\n";
+	  $self->maybe_escapeHTML($text, $q->{'noescape'}), "</a>\n";
     } else {
-	print escapeHTML($text);
+	print $self->maybe_escapeHTML($text, $q->{'noescape'});
     }
     if ($q->{'helpdesc'}) {
-      print "<br><small><i>" . escapeHTML($q->{helpdesc}) . "</i></small>";
+
+      #
+      # Get the actual help text, in case this is a subroutine.
+      #
+      my $helptext = $q->{'helpdesc'};
+      if (ref($helptext) eq "CODE") {
+          $helptext = $helptext->();
+      }
+      print "<br><small><i>" . 
+	$self->maybe_escapeHTML($helptext, $q->{'noescape'}) . "</i></small>";
     }
     print "</td><td" . $self->do_css('qwquestion',$q->{'name'}, 1) . ">\n";
 }
@@ -215,21 +302,25 @@ sub do_question_end {
 
 sub start_questions {
     my ($self, $wiz, $p, $title, $intro) = @_;
+#    print "<td id=\"qwmain\" width=\"100%\">\n";
+    print "<span id=\"qwmain\">\n";
     if ($title) {
 	print $self->{'cgi'}->h1(escapeHTML($title)),"\n";
     }
     if ($intro) {
-	$intro = escapeHTML($intro);
+	$intro = $self->maybe_escapeHTML($intro, $p->{'noescape'});
 	$intro =~ s/\n\n/\n<p class=\"qwintroduction\">\n/g;
 	print "<p class=\"qwintroduction\">$intro\n<p class=\"qwintroduction\">\n";
     }
     print "<table class=\"qwquestions\">\n";
+    $self->{'intable'} = 1;
 }
 
 sub end_questions {
     my ($self, $wiz, $p) = @_;
     print "</table>\n";
-
+#    print "</td>\n";
+    print "</span>\n";
     #
     # This focus() call should allow the user to type directly into the
     # first text box without having to click there first.
@@ -239,6 +330,7 @@ sub end_questions {
     print "</script>\n";
 
     $self->{'started'} = $wiz->{'started'} = 0;
+    delete($self->{'intable'});
 }
 
 sub do_pass {
@@ -247,17 +339,121 @@ sub do_pass {
       if ($self->qwparam($name) ne '');
 }
 
+##################################################
+# Bar support
+##################################################
+
+sub start_bar {
+    my ($self, $wiz, $name) = @_;
+    if ($self->{'intable'}) {
+	print "</table>\n";
+    }
+    print "<div " . $self->do_css('qwbar',$name) . ">\n";
+}
+
+sub end_bar {
+    my ($self, $wiz, $name) = @_;
+    print "</div>\n";
+    if ($self->{'intable'}) {
+	print "<table class=\"qwquestions\">\n";
+    }
+}
+
+sub do_bar {
+    my ($self, $q, $wiz, $p, $widgets) = @_;
+
+    $self->start_bar($wiz, undef);
+    $self->do_a_table([$widgets], 0, $wiz, $q, $p);
+    $self->end_bar($wiz, 'Questions');
+}
+
+sub do_top_bar {
+    my ($self, $q, $wiz, $p, $widgets) = @_;
+
+    print "<tr><td colspan=\"10\">" if (!exists($self->{'cssurl'}));
+    $self->do_a_table([$widgets], 0, $wiz, $q, $p, 'topbar', 'topbar','topbar');
+    print "</td></tr>\n" if (!exists($self->{'cssurl'}));
+}
+
+sub start_center_section {
+    my ($self) = @_;
+    $self->open_span_or_td("centersection");
+}
+
+sub end_center_section {
+    my ($self, $wiz, $p, $next) = @_;
+    print "<div class=\"buttons\">\n";
+    print "  <input class=qwnext type=submit value=\"" . escapeHTML(QWizard::Generator::remove_accelerator($next)) . "\">\n";
+    $self->do_hidden($wiz, "redo_screen", 0) if (!$self->qwparam('redo_screen'));
+    if ($self->qwparam('allow_refresh') || $p->{'allow_refresh'}) {
+	print "<input type=submit onclick=\"$redo_screen_js\" name=redo_screen_but value=\"Refresh Screen\">\n";
+    }
+    print "  <input class=qwcancel name=\"qw_cancel\" type=submit value=\"Cancel\">\n" if (!$self->{'no_cancel'});
+    print "</div>\n";
+    $self->close_span_or_td();
+}
+
+sub start_primaries {
+    my ($self) = @_;
+    # this is closed in wait_for()
+    $self->open_div_or_table("qwizard");
+    print "<tr>\n" if (!exists($self->{'cssurl'}));
+}
+
+sub do_side {
+    my ($self, $spot, $q, $wiz, $p, $widgets) = @_;
+    $self->open_span_or_td($spot);
+
+    my @tableinfo;
+    foreach my $w (@$widgets) {
+	next if (!$w);
+	if (ref($w) eq 'ARRAY') {
+	    # special stand-alone side component
+	    my $title = "";
+	    $title = shift(@$w) if (ref($w->[0]) eq '');
+	    my $id = $title;
+	    $id =~ s/\W//;
+	    print "<div id=\"side$id\" class=\"side\">\n";
+	    print "<div id=\"sidetitle$id\" class=\"sidetitle\">" .
+	      escapeHTML($title) . "</div>\n" if ($title);
+	    print "<table id=\"sidecontent$id\" class=\"sidecontent\">\n";
+	    foreach my $widget (@$w) {
+		# print "widget: $widget\n";
+		$wiz->ask_question($p, $widget);
+	    }
+	    print "</table>\n";
+	    print "</div>\n";
+	} else {
+	    # add to the default (bottom) table component
+	    push @tableinfo, [$w];
+	}
+    }
+
+    $self->do_a_table(\@tableinfo, 0, $wiz, $q, $p, $spot);
+    $self->close_span_or_td();
+}
+
+sub do_left_side {
+    my $self = shift;
+    $self->do_side('leftside', @_);
+}
+
+sub do_right_side {
+    my $self = shift;
+    $self->do_side('rightside', @_);
+}
+
 ##################################################
 # widgets
 ##################################################
 
 sub do_button {
     my ($self, $q, $wiz, $p, $vals) = @_;
-    print "<input" . $self->do_css('qwbutton',$q->{'name'}) . " type=submit name=\"$q->{'name'}\" value=\"" . $vals . "\">\n";
+    print "<input" . $self->do_css('qwbutton',$q->{'name'}) . " type=submit name=\"$q->{'name'}\" value=\"" . QWizard::Generator::remove_accelerator($vals) . "\">\n";
 }
 
 sub do_checkbox {
-    my ($self, $q, $wiz, $p, $vals, $def, $submit) = @_;
+    my ($self, $q, $wiz, $p, $vals, $def, $submit, $refresh_on_change) = @_;
     $vals = [1, 0] if ($#$vals == -1);
     my $otherstuff;
     if ($def == $vals->[0]) {
@@ -269,29 +465,49 @@ sub do_checkbox {
     if ($submit) {
 	$otherstuff .= " onclick=\"this.form.submit()\"";
     }
+    if ($refresh_on_change) {
+	$otherstuff .= " onclick=\"$redo_screen_js\"";
+    }
     print "<input" . $self->do_css('qwcheckbox',$q->{'name'}) . " type=checkbox name=\"$q->{name}\"$otherstuff>";
 }
 
 sub do_multicheckbox {
-    my ($self, $q, $wiz, $p, $defs, $vals, $labels, $submit) = @_;
+    my ($self, $q, $wiz, $p, $defs, $vals, $labels,
+	$submit, $refresh_on_change) = @_;
     print "<table>";
     my $count = -1;
     my ($startname, $endname);
     foreach my $v (@$vals) {
 	$count++;
 	my $otherstuff;
+
+	if ($wiz->qwparam('redoing_now')) {
+	    $otherstuff .= "checked"
+	      if ($wiz->qwparam($q->{'name'} . $v) eq $v);
+	} else {
+	    $otherstuff .= "checked" if ($defs->[$count]);
+	}
+	
+
 	$otherstuff .= "checked" if ($defs->[$count]);
 	if ($submit) {
 	    $otherstuff .= " onclick=\"this.form.submit()";
 	}
-	my $l = (($labels->{$v}) ? $labels->{$v} : "$v");
+	if ($refresh_on_change) {
+	    $otherstuff .= " onclick=\"$redo_screen_js\"";
+	}
+	my $l = QWizard::Generator::remove_accelerator(($labels->{$v}) ? $labels->{$v} : "$v");
 	print "<tr><td>" . escapeHTML($l)  . "</td>\n";
-	print "<td><input" . $self->do_css('qwmulticheckbox',$q->{'name'}) . " $otherstuff value=\"" . escapeHTML($v) . 
-	  "\" type=checkbox name=\"$q->{name}$l\"></td></tr>";
+	print "<td><input" . 
+	  $self->do_css('qwmulticheckbox',$q->{'name'}) . 
+	    " $otherstuff value=\"" . 
+	      escapeHTML($v) .
+		"\" type=checkbox name=\"" .
+		  escapeHTML("$q->{name}$v") . "\"></td></tr>";
 	# XXX: hack:
 	push @{$wiz->{'passvars'}},$q->{'name'} . $v;
-	$startname = "$q->{name}$l" if ($count == 0);
-	$endname = "$q->{name}$l" if ($count == $#$vals);
+	$startname = escapeHTML("$q->{name}$v") if ($count == 0);
+	$endname = escapeHTML("$q->{name}$v") if ($count == $#$vals);
     }
 
     print "</table>";
@@ -371,18 +587,32 @@ sub do_multicheckbox {
 }
 
 sub do_radio {
-    my ($self, $q, $wiz, $p, $vals, $labels, $def, $submit, $name) = @_;
-    my @stuff;
-    push @stuff, -onclick, "this.form.submit()" if ($submit);
+    my ($self, $q, $wiz, $p, $vals, $labels, $def,
+	$submit, $refresh_on_change, $name, $icons, $iconwidth) = @_;
+    my $stuff;
+    $stuff = " onclick=\"this.form.submit()\" " if ($submit);
+    $stuff = " onclick=\"$redo_screen_js\" " if ($refresh_on_change);
 
-    print $self->{'cgi'}->radio_group(-name => $name,
-				      -id => 'qwradio$name',
-				      -class => 'qwradio',
-				      -values => $vals,
-				      -linebreak => 'true',
-				      -labels => $labels,
-				      -default => $def,
-				      @stuff);
+    # remove the key accelerators
+    my %passlabs = %$labels;
+    # remove key bindings specifiers
+    map {
+	$passlabs{$_} = QWizard::Generator::remove_accelerator($passlabs{$_});
+    } keys(%passlabs);
+
+    # correct the ordering
+    my @passvals = reverse @$vals;
+
+    foreach my $value (@passvals) {
+	print "  <input type=\"radio\" name=\"$name\" value=\"" .
+	  escapeHTML($value) . "\" $stuff " .
+	    $self->do_css('qwradio',$q->{'name'}) . " />\n    " .
+	      (($icons->{$value}) ?
+	       $self->get_image($icons->{$value}, 'qwradioimg', $q->{'name'}) :
+	       (($iconwidth) ? "<span style=\"width: ${iconwidth};\" />" : "")).
+		 $passlabs{$value} .
+		   "<br />\n";
+    }
 }
 
 
@@ -417,14 +647,17 @@ sub do_paragraph {
 }
 
 sub do_menu {
-    my ($self, $q, $wiz, $p, $vals, $labels, $def, $submit, $name) = @_;
+    my ($self, $q, $wiz, $p, $vals, $labels, $def,
+	$submit, $refresh_on_change, $name) = @_;
     my @stuff;
     push @stuff, -onchange, "this.form.submit()" if ($submit);
+    push @stuff, -onchange, "$redo_screen_js" if ($refresh_on_change);
 
     print $self->{'cgi'}->popup_menu(-name => $name,
 				     -id => 'qwmenu' . $name,
 				     -class => 'qwmenu',
 				     -values => $vals,
+				     -override => 1,
 				     -labels => $labels,
 				     -default => $def,
 				     @stuff);
@@ -435,8 +668,9 @@ sub do_fileupload {
 
     push @{$wiz->{'passvars'}}, $q->{'name'} . "_qwf";
     print $self->{'cgi'}->filefield(-name => $q->{name},
-				     -id => 'qwmenu' . $q->{'name'},
-				     -class => 'qwmenu',
+				    -id => 'qwmenu' . $q->{'name'},
+				    -class => 'qwmenu',
+				    -override => 1,
 				    -default => $def);
 }
 
@@ -517,7 +751,8 @@ sub qw_upload_fh {
 }
 
 sub do_entry {
-    my ($self, $q, $wiz, $p, $name, $def, $hide, $size, $maxsize, $submit) = @_;
+    my ($self, $q, $wiz, $p, $name, $def, $hide, $size, $maxsize,
+	$submit, $refresh_on_change) = @_;
     my $otherinfo;
     if ($size) {
 	$otherinfo .= " size=\"$size\"";
@@ -535,6 +770,9 @@ sub do_entry {
     if ($submit) {
 	$otherinfo .= " onchange=\"this.form.submit()\"";
     }
+    if ($refresh_on_change) {
+	$otherinfo .= " onclick=\"$redo_screen_js\"";
+    }
 
     #
     # If the hide flag was set, we'll treat this as unprintable text.
@@ -548,7 +786,7 @@ sub do_entry {
 }
 
 sub do_textbox {
-    my ($self, $q, $wiz, $p, $def, $width, $size, $height, $submit) = @_;
+    my ($self, $q, $wiz, $p, $def, $width, $size, $height, $submit, $refresh_on_change) = @_;
     my $otherinfo;
     if ($size || $width) {
 	$size = $size || $width;
@@ -560,14 +798,18 @@ sub do_textbox {
     if ($submit) {
 	$otherinfo .= " onchange=\"this.form.submit()\"";
     }
+    if ($refresh_on_change) {
+	$otherinfo .= " onclick=\"$redo_screen_js\"";
+    }
     print "<textarea" . $self->do_css('qwtextbox',$q->{'name'}) . 
       " name=\"$q->{name}\" $otherinfo>" . escapeHTML($def) . "</textarea>";
 }
 
 sub do_error {
     my ($self, $q, $wiz, $p, $err) = @_;
-    print "<tr" . $self->do_css('qwerrorrow',$q->{'name'}) . "><td" .
-      $self->do_css('qwerrorcol',$q->{'name'}) .
+    my $name = ($q ? $q->{'name'} : '');
+    print "<tr" . $self->do_css('qwerrorrow',$name) . "><td" .
+      $self->do_css('qwerrorcol',$name) .
 	" colspan=3><font color=red>" . escapeHTML($err) .
 	  "</font></td></tr>\n";
 }
@@ -606,7 +848,8 @@ sub do_table {
     my ($self, $q, $wiz, $p, $table, $headers) = @_;
     my $color = $self->{'tablebgcolor'} || $self->{'bgcolor'};
     print "<table" . $self->do_css('qwtable',$q->{'name'}) .
-      " bgcolor=$color border=1>\n";
+      (exists($self->{'cssurl'}) ? "" : "bgcolor=$color border=1>") . 
+       "\n";
 
     if ($headers) {
 	print " <tr " . $self->do_css('qwtableheaderrow',$q->{'name'}) .
@@ -623,17 +866,18 @@ sub do_table {
 }
 
 sub do_a_table {
-    my ($self, $table, $started, $wiz, $q, $p) = @_;
-    print "<table" . $self->do_css('qwsubtable',$q->{'name'}) . ">"
+    my ($self, $table, $started, $wiz, $q, $p, $name) = @_;
+    $name = $q->{'name'} if (!$name);
+    print "<table" . $self->do_css('qwsubtable',$name) . ">"
       if (!$started);
     foreach my $row (@$table) {
-	print " <tr" . $self->do_css('qwtablerow',$q->{'name'}) . ">\n";
+	print " <tr" . $self->do_css('qwtablerow',$name) . ">\n";
 	foreach my $column (@$row) {
 	    print "<td>";
 	    if (ref($column) eq "ARRAY") {
 	        $self->do_a_table($column, 0, $wiz, $q, $p);
 	    } elsif (ref($column) eq "HASH") {
-		print "<table" . $self->do_css('qwtablewidget',$q->{'name'}) .
+		print "<table" . $self->do_css('qwtablewidget',$name) .
 		  ">\n";
 		my $param = $wiz->ask_question($p, $column);
 		push @{$wiz->{'passvars'}}, $param;
@@ -659,7 +903,7 @@ sub do_graph {
 	
 	# XXX: net-policy specific hack!
 	print "<img" . $self->do_css('qwgraph',$q->{'name'}) .
-	  " src=\"?webfile=$file\">\n";
+	  " src=\"" . $self->{'imgpath'} . escapeHTML($file) . "\">\n";
     } else {
 	print "graphs not supported without additional software\n";
     }
@@ -679,7 +923,7 @@ sub do_image {
 	} else {
 	    $image = $imgfile;
 	}
-	my $imagesrc = "src=\"?webfile=$image\"";
+	my $imagesrc = "src=\"" . $self->{'imgpath'} . escapeHTML($image) ."\"";
 
 	#
 	# If an alt tag was specified, create the alt image message.
@@ -752,13 +996,18 @@ EOF
 #TODO: Support passing in a hash for tree data (instead of just a function)
 
 sub do_tree {
-    my ($self, $q, $wiz, $p, $labels) = @_;
+    my ($self, $q, $wiz, $p, $labels, $expand_all, $def) = @_;
 
     my $treename = $q->{'name'} || 'tree';
 
     my $expanded = $self->qwparam("${treename}_expanded") || $q->{'root'};
     my @expand = split(/,/, $expanded);
-    my $redo = $self->qwparam("redo_screen");
+    # redo_screen values:
+    #  1: selects a label
+    #  2: expands a branch
+    #  3: collapses a branch
+    my $redo = $self->qwparam("redoing_now");
+
     if ($redo == 2 && $self->qwparam("${treename}_collapse")) {
 	push @expand, $self->qwparam("${treename}_collapse");
     } elsif ($redo == 3 && $self->qwparam("${treename}_collapse")) {
@@ -781,7 +1030,7 @@ sub do_tree {
 	    }
 	}
     } else { #ensure that the default is initially visible
-	$selected = $q->{'default'} || $q->{'root'} || return;
+	$selected = $def || $q->{'root'} || return;
 	my $cur = $selected;
 	until ($cur eq $q->{'root'}) {
 	    $cur = get_name($q->{'parent'}->($wiz, $cur) || return);
@@ -793,8 +1042,6 @@ sub do_tree {
     $self->do_hidden($wiz, "${treename}_expanded", $expanded);
     if ($self->{'first_tree'}) { #only one hidden value for redo_screen
 	$self->{'first_tree'} = 0;
-	$self->do_hidden($wiz, "redo_screen", '0');
-	$self->qwparam("redo_screen", $redo); #save the value of tree_redo
     }
     $self->do_hidden($wiz, $treename, $selected);
 
@@ -816,7 +1063,10 @@ sub do_tree {
 </script>
 EOF
 
-    print_branch($wiz, $q, $q->{'root'}, $selected, 0, $labels, @expand);
+    print "<div " . $self->do_css('qwtree',$treename) . ">\n";
+    $self->print_branch($wiz, $q, $q->{'root'}, $selected, 0, $labels,
+			\@expand, $expand_all);
+    print "</div>\n";
 }
 
 sub get_name {
@@ -832,35 +1082,37 @@ sub get_name {
 #recursively print out the tree
 sub print_branch {
     # XXX: css this
-    my ($wiz, $q, $cur, $selected, $nest, $labels, @expand) = @_;
+    my ($self, $wiz, $q, $cur, $selected, $nest, $labels,
+	$expand, $expand_all) = @_;
 
     print "<br>" if $nest;
     for my $i (1 .. (5 * $nest)) { print "&nbsp;"; }
 
     my $children = $q->{'children'}->($wiz, get_name($cur));
     if ($#$children > -1) {
-	my @ans = grep($_ eq get_name($cur), @expand); 
-	if ($#ans > -1) { #is it expanded?
-	    make_link('minus', 3, $cur, $selected, $q, $labels);
+	my @ans = grep($_ eq get_name($cur), @$expand); 
+	if ($#ans > -1 || $expand_all > 0) { #is it expanded?
+	    $self->make_link('minus', 3, $cur, $selected, $q, $labels);
 	    foreach my $child (@$children) {
-		print_branch($wiz, $q, $child, $selected, $nest + 1, $labels, @expand);
+		$self->print_branch($wiz, $q, $child, $selected,
+				    $nest + 1, $labels, $expand, $expand_all-1);
 	    }
 	} else {
-	    make_link('plus', 2, $cur, $selected, $q, $labels);
+	    $self->make_link('plus', 2, $cur, $selected, $q, $labels);
 	}
     } else {
-	make_link('blank', 0, $cur, $selected, $q, $labels);
+	$self->make_link('blank', 0, $cur, $selected, $q, $labels);
     }
 }
 
 # prints a single node, and any required links, etc
 sub make_link {
     # XXX: css this
-    my ($imgtype, $oper, $cur, $selected, $q, $labels) = @_;
+    my ($self, $imgtype, $oper, $cur, $selected, $q, $labels) = @_;
     my $name = get_name($cur);
     my $treename = $q->{'name'} || 'tree';
     print "<a href=\"javascript:${treename}_select('$name', $oper)\">" if $oper;
-    print "<img src=\"?webfile=tree_$imgtype.png\" border=0>";
+    print "<img src=\"$self->{'imgpath'}tree_$imgtype.png\" border=0>";
     print "</a>" if $oper;
     print "&nbsp;";
     my $label;
@@ -984,4 +1236,11 @@ sub make_displayable {
     return $str;
 }
 
+sub maybe_escapeHTML {
+    my ($self, $text, $noescapeit) = @_;
+    return $text if ($self->{'noescape'} || $noescapeit);
+    return escapeHTML($text);
+}
+
+
 1;
