@@ -1,7 +1,13 @@
 package QWizard::Generator::Tk;
 
+#
+# TODO:
+#  - better layout; currently ugly
+#  - bar support
+#  - left/right side support
+
 use strict;
-my $VERSION = '2.2';
+my $VERSION = '3.0';
 use Tk;
 use Tk::Table;
 use Tk::Pane;
@@ -19,6 +25,9 @@ sub new {
     my $type = shift;
     my ($class) = ref($type) || $type;
     my $self = {'keep_working_hook' => \&QWizard::Generator::backup_params};
+    for (my $i = 0; $i <= $#_; $i += 2) {
+	$self->{$_[$i]} = $_[$i+1];
+    }
     bless($self, $class);
     $self->add_handler('text',\&QWizard::Generator::Tk::do_entry,
 		       [['single','name'],
@@ -26,7 +35,8 @@ sub new {
 			['forced','0'],
 			['single','size'],
 			['single','maxsize'],
-			['single','submit']]);
+			['single','submit'],
+			['single','refresh_on_change']]);
     # XXX: we need to do a real text box
     $self->add_handler('textbox',\&QWizard::Generator::Tk::do_entry,
 		       [['single','name'],
@@ -34,18 +44,21 @@ sub new {
 			['forced','0'],
 			['single','size'],
 			['single','maxsize'],
-			['single','submit']]);
+			['single','submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('hidetext',\&QWizard::Generator::Tk::do_entry,
 		       [['single','name'],
 			['default'],
 			['forced','1'],
 			['single','size'],
 			['single','maxsize'],
-			['single','submit']]);
+			['single','submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('checkbox',\&QWizard::Generator::Tk::do_checkbox,
 		       [['multi','values'],
 			['default'],
-			['single', 'submit']]);
+			['single', 'submit'],
+			['single','refresh_on_change']]);
     $self->add_handler('multi_checkbox',
 		       \&QWizard::Generator::Tk::do_multicheckbox,
 		       [['multi','default'],
@@ -106,12 +119,19 @@ sub goto_top {
 
 sub goto_next {
     shift if (ref($_[0]) ne 'QWizard::Generator::Tk');
-    my ($self, $ignorefirst, $val) = @_;
-    if ($ignorefirst && ref($ignorefirst) eq 'SCALAR' && $$ignorefirst) {
-	$$ignorefirst--;
+    my ($self, $ignorefirst_or_varname, $refresh_on_change, $val) = @_;
+    if ($ignorefirst_or_varname &&
+	ref($ignorefirst_or_varname) eq 'SCALAR' && $$ignorefirst_or_varname) {
+	$$ignorefirst_or_varname--;
 	return;
-    } elsif ($ignorefirst && ref($ignorefirst) ne 'SCALAR') {
-	$self->qwparam($ignorefirst, $val);
+    } elsif (ref($ignorefirst_or_varname) ne 'SCALAR') {
+	if ($ignorefirst_or_varname) {
+	    $self->qwparam($ignorefirst_or_varname, $val);
+	}
+    }
+    if ($refresh_on_change) {
+	print STDERR "redo:!!!\n";
+	$self->qwparam('redo_screen',1);
     }
 
     $self->unmake_top();
@@ -309,19 +329,29 @@ sub do_error {
 sub do_question {
     my ($self, $q, $wiz, $p, $text, $qcount) = @_;
     my $top = $self->{'qtable'};
+
     $self->{'currentq'} = $qcount + $self->{'qadd'};
-    if ($q->{'helpdesc'} && !$self->qwpref('usehelpballons')) {
+
+    #
+    # Get the actual help text, in case this is a subroutine.
+    #
+    my $helptext = $q->{'helpdesc'};
+    if (ref($helptext) eq "CODE") {
+	$helptext = $helptext->();
+    }
+
+    if ($helptext && !$self->qwpref('usehelpballons')) {
 	my $f = $top->Frame();
 	$f->Label(-text => $text, -anchor => 'nw')->pack(-anchor => 'w');
-	$f->Label(-text => $q->{'helpdesc'}, -anchor => 'nw',
+	$f->Label(-text => $helptext, -anchor => 'nw',
 		  -font => 'Helvetica 12 italic')
 	  ->pack(-anchor => 'w');
 	$self->put_it($f, undef, 1);
     } else {
 	my $l = $top->Label(-text => $text, -anchor => 'nw');
 	$self->put_it($l, undef, 1);
-	if ($self->{'balloon'} && $q->{'helpdesc'}) {
-	    $self->{'balloon'}->attach($l, -balloonmsg => $q->{'helpdesc'});
+	if ($self->{'balloon'} && $helptext) {
+	    $self->{'balloon'}->attach($l, -balloonmsg => $helptext);
 	    # XXX: change the "help" window text, which doesn't exist yet.
 	}
     }
@@ -359,7 +389,8 @@ sub get_extra_args {
 
     my @args;
 
-    if ($q->{'submit'} && $q->{type} ne 'text') {
+    if (($q->{'submit'} || $q->{'refresh_on_change'}) &&
+      $q->{type} ne 'text') {
 	my $ignorefirst = 0;
 	if ($q->{'type'} eq 'menu') {
 	    # menus do an initial call immediately after being created.
@@ -367,7 +398,8 @@ sub get_extra_args {
 	    # (did I mention "sigh"?)
 	    $ignorefirst = 1;
 	}
-	push @args, '-command', [\&goto_next, $self, \$ignorefirst];
+	push @args, '-command', [\&goto_next, $self, \$ignorefirst,
+				 $q->{'refresh_on_change'}];
     }
     return \@args;
 }
@@ -389,6 +421,7 @@ sub do_button {
     my $but = $self->{'qtable'}->Button(-text => $vals,
 					-command => [\&goto_next, 
 						     $self, $q->{'name'},
+						     $q->{'refresh_on_change'},
 						     $def]);
     $self->put_it($but);
 }
@@ -656,12 +689,14 @@ sub do_image {
 	    if ($datastr) {
 		require MIME::Base64;
 		$ph = $self->{'qtable'}->Photo(
-					       -data => 
+					       -format => 'png',
+					       -data =>
 					       MIME::Base64::encode_base64($datastr));
 
 	    } else {
 		# image file
-		$ph = $self->{'qtable'}->Photo(-file => $wiz->{'generator'}{'imagebase'} . $filestr);
+		$ph = $self->{'qtable'}->Photo(-format => 'png',
+					       -file => $wiz->{'generator'}{'imagebase'} . $filestr);
 	    }
 	}
 	if ($ph) {
@@ -739,7 +774,7 @@ sub add_node {
     my $child = $tree->addchild($parent, -text => $label,
 				-data => $name);
     my $ans = $q->{'children'}->($wiz, $node);
-    $tree->setmode($child, $#$ans > -1 ? 'open' : 'none');
+    $tree->setmode($child, ($ans && $#$ans > -1) ? 'open' : 'none');
 
     $tree->selectionSet($child) if ($name eq $q->{'default'});
     if ($name eq $exp) {
